/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"

typedef struct run_s
{
  int start, end, label;
} run_t;

/* This should be done on the GPU */
void rle(CuImage &img, run_t **run, int *runs, int **row, int *rows)
{
  int alloced = 256;

  *run = (run_t*)malloc(alloced*sizeof(run_t));
  *runs = 0;

  int dx = img.GetDimensions().Dim(0),
      dy = img.GetDimensions().Dim(1);

  *row = (int*)malloc(dy*sizeof(int));
  *rows = dy;

  unsigned char *data = img.GetDataInt8();

  for (int yy=0; yy < dy; yy++)
  {
    for (int xx=0; xx < dx; xx++)
    {
      if (data[yy*dx+xx])
      {
        if ((*runs) == alloced)
        {
          alloced = alloced * 2;
          *run = (run_t*)realloc(*run, alloced*sizeof(run_t));
        }

        (*run)[(*runs)].start = xx;
        for (; xx < dx && data[yy*dx+xx]; xx++);
        (*run)[(*runs)].end = xx-1;
        (*run)[(*runs)].label = (*runs)++;
      }
    }

    (*row)[yy] = *runs;
  }

  free(data);
}

/* run is inout parameter with prelabeled runs (i.e. each run has
 * a unique label. The labels must be regularly increasing from
 * left to right, top to bottom
 */
void cca(run_t *run, int runs, int *row, int rows)
{
  int *eq = (int*) malloc(runs*sizeof(int));

  /* Initialize equivalences */
  for (int ii=0; ii < runs; ii++)
    eq[ii] = ii;

  /* Connect runs, noting equivalences */
  for (int yy=2; yy < rows; yy++)
  {
    for (int xx=row[yy-1]; xx < row[yy]; xx++)
    {
      int minrun = xx, start = runs, end = -1;

      /* Find minimum label of connecting previous runs */
      for (int px=row[yy-2]; px < row[yy-1]; px++)
        if (run[px].end >= run[xx].start && run[px].start <= run[xx].end)
        {
          if (start == runs) start = px;
          minrun = min(minrun, eq[run[px].label]);
          end = px;
        }

      /* Set equivalences of connecting previous runs to the minimum label */
      for (int px=start; px <= end; px++)
        eq[run[px].label] = minrun;

      /* Label current run with the minimum of connecting runs */
      run[xx].label = minrun;
    }
  }

  /* Resolve and apply equivalences */
  for (int ii=0; ii < runs; ii++)
  {
    eq[ii] = eq[eq[ii]];
    run[ii].label = eq[run[ii].label];
  }

  /* Refine label numbers */
  for (int ii=0; ii < runs; ii++)
    eq[run[ii].label] = -1;

  int label = 1;
  for (int ii=0; ii < runs; ii++)
    if (eq[ii] == -1)
      eq[ii] = label++;

  for (int ii=0; ii < runs; ii++)
    run[ii].label = eq[run[ii].label];
}

CuImage label(CuImage &img)
{
  run_t *run;
  int runs, *row, rows;
  int dx = img.GetDimensions().Dim(0);
  unsigned short *data = (unsigned short*) calloc(img.GetDimensions().Size(), sizeof(unsigned short));

  CuImage res(img.GetDimensions());

  rle(img, &run, &runs, &row, &rows);
  cca(run, runs, row, rows);

  for (int yy=1; yy < rows; yy++)
    for (int xx=row[yy-1]; xx < row[yy]; xx++)
      for (int pp=run[xx].start; pp <= run[xx].end; pp++)
        data[yy*dx+pp] = run[xx].label;

  res.SetDataInt16(data);

  free(data);
  free(run);
  free(row);

  return res;
}
