/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"

__global__ void CuImageFFTFloatToComplex
(   float *in, cuComplex *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
    {
			out[ii].x = in[ii];
			out[ii].y = 0;
    }
}

__global__ void CuImageFFTDeinterleave
(   cuComplex *in, float *real, float *imag, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
    {
			real[ii] = in[ii].x;
			imag[ii] = in[ii].y;
    }
}

__global__ void CuImageFFTComplexToFloat
(   cuComplex *in, float *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = in[ii].x;
}

__global__ void CuImageFFTInterleave
(   float *real, float *imag, cuComplex *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
    {
			out[ii].x = real[ii];
			out[ii].y = imag[ii];
    }
}

void fft(CuImage &img, CuImage &real, CuImage &imag)
{
  real = CuImage(img.GetDimensions());
  imag = CuImage(img.GetDimensions());

  CuImage complex = CuImage(cmmDims(img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), 2));
	cufftHandle plan;
	cufftResult result;

  /* Make a plan */
	result = cufftPlan2d(&plan, img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), CUFFT_C2C);
  if (result != CUFFT_SUCCESS)
		return;

  /* Set up threading */
  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(img.GetDimensions().Size(), CUIMAGE_THREADS));

  /* Convert real input data to complex */
  complex.Lock();
  img.Lock();
  CuImageFFTFloatToComplex<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), (cufftComplex*)complex.GetDevicePtr(), img.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  img.Release();
  
  /* Do the actual FFT */
	CUFFT_SAFE_CALL( cufftExecC2C(plan, (cufftComplex*)complex.GetDevicePtr(), (cufftComplex*)complex.GetDevicePtr(), CUFFT_FORWARD) );

  /* Deinterleave complex data */
  real.Lock();
  imag.Lock();
  CuImageFFTDeinterleave<<<blockGrid, threadBlock>>>((cufftComplex*)complex.GetDevicePtr(), real.GetDevicePtr(), imag.GetDevicePtr(), img.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  imag.Release();
  real.Release();
  complex.Release();

  cufftDestroy(plan);
}

CuImage ifft(CuImage &real, CuImage &imag)
{
  CuImage img = CuImage(real.GetDimensions());

  CuImage complex = CuImage(cmmDims(img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), 2));
	cufftHandle plan;
	cufftResult result;

  /* Make a plan */
	result = cufftPlan2d(&plan, img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), CUFFT_C2C);
  if (result != CUFFT_SUCCESS)
		return img;

  /* Set up threading */
  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(img.GetDimensions().Size(), CUIMAGE_THREADS));

  /* Interleave complex data */
  complex.Lock();
  real.Lock();
  imag.Lock();
  CuImageFFTInterleave<<<blockGrid, threadBlock>>>(real.GetDevicePtr(), imag.GetDevicePtr(), (cufftComplex*)complex.GetDevicePtr(), img.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  imag.Release();
  real.Release();
  
  /* Do the actual FFT */
	CUFFT_SAFE_CALL( cufftExecC2C(plan, (cufftComplex*)complex.GetDevicePtr(), (cufftComplex*)complex.GetDevicePtr(), CUFFT_INVERSE) );

  /* Get real part of complex data */
  img.Lock();
  CuImageFFTComplexToFloat<<<blockGrid, threadBlock>>>((cufftComplex*)complex.GetDevicePtr(), img.GetDevicePtr(), img.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  img.Release();
  complex.Release();

  cufftDestroy(plan);

  return img;
}
