/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"



/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  Users and possessors of this source code 
 * are hereby granted a nonexclusive, royalty-free license to use this code 
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.   This source code is a "commercial item" as 
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer  software"  and "commercial computer software 
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein. 
 *
 * Any use of this source code in individual and commercial software must 
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

//Total number of possible data values
#define      BIN_COUNT 256
#define HISTOGRAM_SIZE (BIN_COUNT * sizeof(unsigned int))

//Machine warp size
#ifndef __DEVICE_EMULATION__
//G80's warp size is 32 threads
#define WARP_LOG_SIZE 5
#else
//Emulation currently doesn't execute threads in coherent groups of 32 threads,
//which effectively means warp size of 1 thread for emulation modes
#define WARP_LOG_SIZE 0
#endif


//Warps in thread block
#define  WARP_N 6
//Threads per block count
#define     THREAD_N (WARP_N << WARP_LOG_SIZE)
//Per-block number of elements in histograms
#define BLOCK_MEMORY (WARP_N * BIN_COUNT)


#define IMUL(a, b) __mul24(a, b)


__device__ void addData256(volatile unsigned int *s_WarpHist, unsigned int data, unsigned int threadTag){
    unsigned int count;

    do{
        count = s_WarpHist[data] & 0x07FFFFFFU;
        count = threadTag | (count + 1);
        s_WarpHist[data] = count;
    }while(s_WarpHist[data] != count);
}

__global__ void histogram256Kernel(unsigned int *d_Result, unsigned int *d_Data, int dataN){
    //Current global thread index
    const int    globalTid = IMUL(blockIdx.x, blockDim.x) + threadIdx.x;
    //Total number of threads in the compute grid
    const int   numThreads = IMUL(blockDim.x, gridDim.x);
    //Thread tag for addPixel()
    #ifndef __DEVICE_EMULATION__
        //WARP_LOG_SIZE higher bits of counter values are tagged 
        //by lower WARP_LOG_SIZE threadID bits
        const unsigned int threadTag = threadIdx.x << (32 - WARP_LOG_SIZE);
    #else
        //Explicitly set to zero to avoid potential troubles
        const unsigned int threadTag = 0;
    #endif
    //Shared memory cache for each warp in current thread block
    //Declare as volatile to prevent incorrect compiler optimizations in addPixel()
    volatile __shared__ unsigned int s_Hist[BLOCK_MEMORY];
    //Current warp shared memory frame
    const int warpBase = IMUL(threadIdx.x >> WARP_LOG_SIZE, BIN_COUNT);

    //Clear shared memory buffer for current thread block before processing
    for(int pos = threadIdx.x; pos < BLOCK_MEMORY; pos += blockDim.x)
       s_Hist[pos] = 0;

    __syncthreads();
    //Cycle through the entire data set, update subhistograms for each warp
    //Since threads in warps always execute the same instruction,
    //we are safe with the addPixel trick
    for(int pos = globalTid; pos < dataN; pos += numThreads){
        unsigned int data4 = d_Data[pos];

        addData256(s_Hist + warpBase, (data4 >>  0) & 0xFFU, threadTag);
        addData256(s_Hist + warpBase, (data4 >>  8) & 0xFFU, threadTag);
        addData256(s_Hist + warpBase, (data4 >> 16) & 0xFFU, threadTag);
        addData256(s_Hist + warpBase, (data4 >> 24) & 0xFFU, threadTag);
    }

    __syncthreads();
    //Merge per-warp histograms into per-block and write to global memory
    for(int pos = threadIdx.x; pos < BIN_COUNT; pos += blockDim.x){
        unsigned int sum = 0;

        for(int base = 0; base < BLOCK_MEMORY; base += BIN_COUNT)
            sum += s_Hist[base + pos] & 0x07FFFFFFU;

#ifdef ATOMICS
        atomicAdd(d_Result + pos, sum);
#else
        d_Result[IMUL(BIN_COUNT, blockIdx.x) + pos] = sum;
#endif

    }
}


//Thread block (== subhistogram) count
#define BLOCK_N 64


///////////////////////////////////////////////////////////////////////////////
// Merge BLOCK_N subhistograms of BIN_COUNT bins into final histogram
////////////////////////////////////////////////////////////////////////////////
__global__ void mergeHistogram256Kernel(unsigned int *d_Result){
    //gridDim.x   == BIN_COUNT
    //blockDim.x  == BLOCK_N
    //blockIdx.x  == bin counter processed by current block
    //threadIdx.x == subhistogram index
    __shared__ unsigned int data[BLOCK_N];

    //Reads are uncoalesced, but this final stage takes
    //only a fraction of total processing time
    data[threadIdx.x] = d_Result[IMUL(threadIdx.x, BIN_COUNT) + blockIdx.x];

    for(int stride = BLOCK_N / 2; stride > 0; stride >>= 1){
        __syncthreads();
        if(threadIdx.x < stride)
            data[threadIdx.x] += data[threadIdx.x + stride];
    }

    if(threadIdx.x == 0)
        d_Result[blockIdx.x] = data[0];
}



////////////////////////////////////////////////////////////////////////////////
// Put all kernels together
////////////////////////////////////////////////////////////////////////////////
//histogram256kernel() results buffer
unsigned int *d_Result256;

//Internal memory allocation
void initHistogram256(void){
    #ifdef ATOMICS
        CUDA_SAFE_CALL( cudaMalloc((void **)&d_Result256, HISTOGRAM_SIZE ) );
    #else
        CUDA_SAFE_CALL( cudaMalloc((void **)&d_Result256, BLOCK_N * HISTOGRAM_SIZE) );
    #endif
}

//Internal memory deallocation
void closeHistogram256(void){
    CUDA_SAFE_CALL( cudaFree(d_Result256) );
}

//histogram256 CPU front-end
void histogram256GPU(
    unsigned int *h_Result,
    unsigned char *d_Data,
    int dataN
){
    #ifdef ATOMICS
        CUDA_SAFE_CALL( cudaMemset(d_Result256, 0, HISTOGRAM_SIZE) );

        histogram256Kernel<<<BLOCK_N, THREAD_N>>>(
            d_Result256,
            (unsigned int *)d_Data,
            dataN / 4
        );
        CUT_CHECK_ERROR("histogram256Kernel() execution failed\n");

    #else
        histogram256Kernel<<<BLOCK_N, THREAD_N>>>(
            d_Result256,
            (unsigned int *)d_Data,
            dataN / 4
        );
        CUIMAGE_SAFE_CALL(cudaThreadSynchronize());
        CUT_CHECK_ERROR("histogram256Kernel() execution failed\n");

        mergeHistogram256Kernel<<<BIN_COUNT, BLOCK_N>>>(d_Result256);
        CUT_CHECK_ERROR("mergeHistogram256Kernel() execution failed\n");
    #endif

    CUDA_SAFE_CALL( cudaMemcpy(h_Result, d_Result256, HISTOGRAM_SIZE, cudaMemcpyDeviceToHost) );
}

void histogram256(CuImage &img, unsigned int *out)
{
  int first = 1;
  if (first)
  {
    // We don't bother closing the histogram, as it only takes 32k memory
    initHistogram256();
    first = 0;
  }

  unsigned char *chardata;

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(img.GetDimensions().Size(), CUIMAGE_THREADS));

  CUIMAGE_SAFE_CALL(CuImageMalloc((void **)&chardata, img.GetDimensions().Size()*sizeof(unsigned char)));

  img.Lock();

  CuImagePixelFloatToInt8<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), chardata, img.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  img.Release();

  histogram256GPU(out, chardata, img.GetDimensions().Size());

  CUIMAGE_SAFE_CALL(CuImageFree(chardata));
}
