/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"

/* *********************************************************************** *
 * *** Device code ******************************************************* *
 * *********************************************************************** */

#define MAXKERNELRADIUS     7
#define MAXKERNELSIZE (2*MAXKERNELRADIUS+1)*(2*MAXKERNELRADIUS+1)

__device__ __constant__ float d_kernel[MAXKERNELSIZE];

/* *********************************************************************** *
 * *** Functions ********************************************************* *
 * *********************************************************************** */

#define CuImageN2P_I2I(NAME, OP)\
__global__ void NAME\
(   float *in, float *out, int width, int height, int xradius, int yradius\
){\
    const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
    const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;\
\
    if (x > xradius-1 && x < width-xradius &&\
        y > yradius-1 && y < height-yradius)\
      out[IMUL(y, width)+x] = OP<true>(in, x, y, width, height);\
    else if (x < width && y < height)\
      out[IMUL(y, width)+x] = OP<false>(in, x, y, width, height);\
}

#define CuImageN2P_ID2I(NAME, OP)\
__global__ void NAME\
(   float *in, int d, float *out, int width, int height, int xradius, int yradius\
){\
    const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
    const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;\
\
    if (x > xradius-1 && x < width-xradius &&\
        y > yradius-1 && y < height-yradius)\
      out[IMUL(y, width)+x] = OP<true>(in, d, x, y, width, height);\
    else if (x < width && y < height)\
      out[IMUL(y, width)+x] = OP<false>(in, d, x, y, width, height);\
}

#define CuImageN2P_IDD2I(NAME, OP)\
__global__ void NAME\
(   float *in, int d1, int d2, float *out, int width, int height, int xradius, int yradius\
){\
    const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
    const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;\
\
    if (x > xradius-1 && x < width-xradius &&\
        y > yradius-1 && y < height-yradius)\
      out[IMUL(y, width)+x] = OP<true>(in, d1, d2, x, y, width, height);\
    else if (x < width && y < height)\
      out[IMUL(y, width)+x] = OP<false>(in, d1, d2, x, y, width, height);\
}

/* *********************************************************************** *
 * *** Operators ********************************************************* *
 * *********************************************************************** */

template<bool fast>
__device__ float CuImageErosion4(float *in, int x, int y, int width, int height)
{
  float out=in[mirror2D<fast>(x, y, width, height)];

  out = min(out, in[mirror2D<fast>(x, y-1, width, height)]);
  out = min(out, in[mirror2D<fast>(x-1, y, width, height)]);
  out = min(out, in[mirror2D<fast>(x+1, y, width, height)]);
  out = min(out, in[mirror2D<fast>(x, y+1, width, height)]);

  return out;
}

CuImageN2P_I2I(CuImageErosion4I2I, CuImageErosion4)

template<bool fast>
__device__ float CuImageErosion(float *in, int x, int y, int width, int height)
{
  float out=1e38;

  for (int ky=-1; ky < 2; ky++)
    for (int kx=-1; kx < 2; kx++)
      out = min(out, in[mirror2D<fast>(x+kx, y+ky, width, height)]);

  return out;
}

CuImageN2P_I2I(CuImageErosionI2I, CuImageErosion)

template<bool fast>
__device__ float CuImageDilation4(float *in, int x, int y, int width, int height)
{
  float out=in[mirror2D<fast>(x, y, width, height)];

  out = max(out, in[mirror2D<fast>(x, y-1, width, height)]);
  out = max(out, in[mirror2D<fast>(x-1, y, width, height)]);
  out = max(out, in[mirror2D<fast>(x+1, y, width, height)]);
  out = max(out, in[mirror2D<fast>(x, y+1, width, height)]);

  return out;
}

CuImageN2P_I2I(CuImageDilation4I2I, CuImageDilation4)

template<bool fast>
__device__ float CuImageDilation(float *in, int x, int y, int width, int height)
{
  float out=-1e38;

  for (int ky=-1; ky < 2; ky++)
    for (int kx=-1; kx < 2; kx++)
      out = max(out, in[mirror2D<fast>(x+kx, y+ky, width, height)]);

  return out;
}

CuImageN2P_I2I(CuImageDilationI2I, CuImageDilation)

template<bool fast>
__device__ float CuImageConvolveSquare(float *in, int radius, int x, int y, int width, int height)
{
  float out=0;
  const int diameter = (radius<<1)+1;

  for (int ky=-radius; ky < radius+1; ky++)
    for (int kx=-radius; kx < radius+1; kx++)
      out += in[mirror2D<fast>(x+kx, y+ky, width, height)] *
             d_kernel[IMUL((ky+radius), diameter) + kx+radius];

  return out;
}

CuImageN2P_ID2I(CuImageConvolveID2I, CuImageConvolveSquare)

template<bool fast>
__device__ float CuImageConvolve(float *in, int xradius, int yradius, int x, int y, int width, int height)
{
  float out=0;
  const int xdiameter = (xradius<<1)+1;

  for (int ky=-yradius; ky < yradius+1; ky++)
    for (int kx=-xradius; kx < xradius+1; kx++)
      out += in[mirror2D<fast>(x+kx, y+ky, width, height)] *
             d_kernel[IMUL((ky+yradius), xdiameter) + kx+xradius];

  return out;
}

CuImageN2P_IDD2I(CuImageConvolveIDD2I, CuImageConvolve)

/* *********************************************************************** *
 * *** Host code ********************************************************* *
 * *********************************************************************** */

/* *********************************************************************** *
 * *** Functions ********************************************************* *
 * *********************************************************************** */

#define CuImageN2P_I2I_Host(NAME, OP, XRADIUS, YRADIUS)\
CuImage NAME(CuImage &img)\
{\
  CuImage res(img.GetDimensions());\
\
  dim3 threadBlock(16, 12);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));\
\
  img.Lock();\
  res.Lock();\
  OP<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), XRADIUS, YRADIUS);\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  img.Release();\
  res.Release();\
\
  return res;\
}

/* *********************************************************************** *
 * *** Operators ********************************************************* *
 * *********************************************************************** */

CuImage erosion(CuImage &img, int iterations, int connectivity)
{
  CuImage res(img.GetDimensions()), res2(img.GetDimensions());

  dim3 threadBlock(16, 12);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  if (iterations == 1)
  {
    img.Lock();
    res.Lock();

    if (connectivity == 2)
       CuImageErosionI2I<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), 1, 1);
    else
       CuImageErosion4I2I<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), 1, 1);

    CUIMAGE_SAFE_CALL(cudaGetLastError());
    img.Release();
    res.Release();

    return res;
  }
  else
  {
    img.Lock();
    res.Lock();
    res2.Lock();

    CuImage *from = &img, *to = &res;

    for (int ii=0; ii < iterations; ii++)
    {
      switch (connectivity)
      {
        case 1:
          CuImageErosion4I2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          break;
        case 2:
          CuImageErosionI2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          break;
        default:
          if ((ii%2) == 0)
            CuImageErosion4I2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          else
            CuImageErosionI2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
         break;
      }

      if ((ii%2) == 0)
      {
        from = &res;
        to   = &res2;
      }
      else
      {
        from = &res2;
        to   = &res;
      }
    }

    CUIMAGE_SAFE_CALL(cudaGetLastError());
    img.Release();
    res.Release();
    res2.Release();

    return *from;
  }
}

CuImage dilation(CuImage &img, int iterations, int connectivity)
{
  CuImage res(img.GetDimensions()), res2(img.GetDimensions());

  dim3 threadBlock(16, 12);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  if (iterations == 1)
  {
    img.Lock();
    res.Lock();

    if (connectivity == 2)
       CuImageDilationI2I<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), 1, 1);
    else
       CuImageDilation4I2I<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), 1, 1);

    CUIMAGE_SAFE_CALL(cudaGetLastError());
    img.Release();
    res.Release();

    return res;
  }
  else
  {
    img.Lock();
    res.Lock();
    res2.Lock();

    CuImage *from = &img, *to = &res;

    for (int ii=0; ii < iterations; ii++)
    {
      switch (connectivity)
      {
        case 1:
          CuImageDilation4I2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          break;
        case 2:
          CuImageDilationI2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          break;
        default:
          if ((ii%2) == 0)
            CuImageDilation4I2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
          else
            CuImageDilationI2I<<<blockGrid, threadBlock>>>(from->GetDevicePtr(), to->GetDevicePtr(), to->GetDimensions().Dim(0), to->GetDimensions().Dim(1), 1, 1);
         break;
      }

      if ((ii%2) == 0)
      {
        from = &res;
        to   = &res2;
      }
      else
      {
        from = &res2;
        to   = &res;
      }
    }

    CUIMAGE_SAFE_CALL(cudaGetLastError());
    img.Release();
    res.Release();
    res2.Release();

    return *from;
  }
}

CuImage gauss(CuImage &img, float sigma)
{
  float kernel[MAXKERNELSIZE];
  float sum;

  const int radius   = fminf(fmaxf(2*sigma, 1), MAXKERNELRADIUS);
  const int diameter = 2*radius+1;

  sum = 0;
  for (int ky=-radius; ky < radius+1; ky++)
    for (int kx=-radius; kx < radius+1; kx++)
      sum += kernel[(ky+radius)*diameter + kx+radius] = exp(-(kx*kx)/(2*sigma*sigma)
                                                            -(ky*ky)/(2*sigma*sigma));

  for (int ii=0; ii < diameter*diameter; ii++)
    kernel[ii] /= sum;

  return convolve(img, kernel, radius, radius);
}

CuImage gaussx(CuImage &img, float sigma)
{
  float kernel[MAXKERNELSIZE];
  float sum;

  const int radius   = fminf(fmaxf(2*sigma, 1), MAXKERNELRADIUS);
  const int diameter = 2*radius+1;

  sum = 0;
  for (int ky=-radius; ky < radius+1; ky++)
    for (int kx=-radius; kx < radius+1; kx++)
      sum += abs(kernel[(ky+radius)*diameter + kx+radius] = -2*kx/(2*sigma*sigma)*exp(-(kx*kx)/(2*sigma*sigma)
                                                                                      -(ky*ky)/(2*sigma*sigma)));

  for (int ii=0; ii < diameter*diameter; ii++)
    kernel[ii] /= sum;

  return convolve(img, kernel, radius, radius);
}

CuImage gaussy(CuImage &img, float sigma)
{
  float kernel[MAXKERNELSIZE];
  float sum;

  const int radius   = fminf(fmaxf(2*sigma, 1), MAXKERNELRADIUS);
  const int diameter = 2*radius+1;

  sum = 0;
  for (int ky=-radius; ky < radius+1; ky++)
    for (int kx=-radius; kx < radius+1; kx++)
      sum += abs(kernel[(ky+radius)*diameter + kx+radius] = -2*ky/(2*sigma*sigma)*exp(-(kx*kx)/(2*sigma*sigma)
                                                                                      -(ky*ky)/(2*sigma*sigma)));

  for (int ii=0; ii < diameter*diameter; ii++)
    kernel[ii] /= sum;

  return convolve(img, kernel, radius, radius);
}

CuImage convolve(CuImage &img, float *kernel, int xradius, int yradius)
{
  CUIMAGE_SAFE_CALL(cudaMemcpyToSymbol(d_kernel, kernel, (xradius*2+1)*(yradius*2+1)*sizeof(float)));

  CuImage res(img.GetDimensions());

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  img.Lock();
  res.Lock();

  CuImageConvolveIDD2I<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), xradius, yradius, res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1), xradius, yradius);
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  img.Release();
  res.Release();

  return res;
}
