/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define CUIMAGE_THREADS 256
#define CUIMAGE_REDUCE_THREADS 128
#define CUIMAGE_REDUCE_BLOCKS  64
#define CUIMAGE_CHUNKS 16

typedef struct CuImageChunkS
{
  int size;
  void *data;
} CuImageChunkT;

typedef struct CuImageChunkListS
{
  CuImageChunkT *data[CUIMAGE_CHUNKS];
  int length;
} CuImageChunkListT;

static CuImageChunkT chunks[CUIMAGE_CHUNKS];
static CuImageChunkListT freechunks, cachedchunks, usedchunks;

static int numchunks=0;

#define CUIMAGE_SAFE_CALL(v) do {\
                               cudaError_t e = v;\
                               if (e != cudaSuccess) {\
                                 char buf[1024];\
                                 sprintf(buf, "%s:%d : %s\nFunction: %s", __FILE__, __LINE__, cudaGetErrorString(e), #v);\
                                 MessageBoxA(NULL, buf, "CUIMAGE error", MB_OK);\
                                 exit(1);\
                               }\
                             } while (0)

int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

//Align a to nearest higher multiple of b
int iAlignUp(int a, int b){
    return (a % b != 0) ?  (a - a % b + b) : a;
}

static void movechunk(CuImageChunkListT *from, int item, CuImageChunkListT *to)
{
  to->data[to->length++] = from->data[item];

  for (int ii=item; ii < from->length-1; ii++)
    from->data[ii] = from->data[ii+1];
  from->data[from->length-1] = NULL;

  from->length--;
}

cudaError_t CuImageMalloc(void **data, int size)
{
  static int initialized = false;

  *data = NULL;

  if (!initialized)
  {
    for (int ii=0; ii < CUIMAGE_CHUNKS; ii++)
    {
      freechunks.data[ii] = &chunks[ii];
    }
    freechunks.length = CUIMAGE_CHUNKS;
    cachedchunks.length = 0;
    usedchunks.length = 0;

    initialized = true;
  }

  /* Try to find chunk in cached chunks list */
  for (int ii=0; ii < cachedchunks.length; ii++)
    if (cachedchunks.data[ii]->size == size)
    {
      *data = cachedchunks.data[ii]->data;

      /* Move to used chunks list */
      movechunk(&cachedchunks, ii, &usedchunks);
      return cudaSuccess;
    }

  if (!freechunks.length)
  {
    /* Free up an old chunk */
    if (cachedchunks.length)
    {
      /* Free the oldest cached chunk */
      CUIMAGE_SAFE_CALL(cudaFree(cachedchunks.data[0]->data));
      movechunk(&cachedchunks, 0, &freechunks);
    }
    else
    {
      /* Unmanage the oldest used chunk */
      movechunk(&usedchunks, 0, &freechunks);
    }
  }

  /* Allocate a new chunk */
  while (cudaMalloc(data, size) == cudaErrorMemoryAllocation &&
         cachedchunks.length)
  {
    /* Free the oldest cached chunk */
    CUIMAGE_SAFE_CALL(cudaFree(cachedchunks.data[0]->data));
    movechunk(&cachedchunks, 0, &freechunks);
  }

  if (!*data)
    return cudaErrorMemoryAllocation;

  freechunks.data[freechunks.length-1]->data = *data;
  freechunks.data[freechunks.length-1]->size = size;

  /* Assign to used chunks list */
  movechunk(&freechunks, freechunks.length-1, &usedchunks);

  return cudaSuccess;
}

cudaError_t CuImageFree(void *data)
{
  /* Try to find data in used chunks list */
  for (int ii=0; ii < usedchunks.length; ii++)
  {
    if (usedchunks.data[ii]->data == data)
    {
      /* Move to cached chunks list */
      movechunk(&usedchunks, ii, &cachedchunks);
      return cudaSuccess;
    }
  }

  /* Data was unmanaged, free immediately */
  CUIMAGE_SAFE_CALL(cudaFree(data));

  return cudaSuccess;
}
