/*
>> cmm.cu
>>
>> Memory manager for CUDA image processing
>> (c) 2007 Wouter Caarls, MPIbpc
*/

#include <cuda.h>

#include "cmm.h"

using namespace cmm;

static cmmState cmmStateNascent = cmmState(new cmmPageList());
static cmmState cmmStateStored  = cmmState(new cmmPageList());
static cmmState cmmStateLocked  = cmmState(new cmmPageList());
static cmmState cmmStateCached  = cmmState(new cmmPageList());
static cmmState cmmStateZombie  = cmmState(new cmmPageList());

static bool initialized = false;

static int page_locks=0, page_relocks=0, page_hits=0, page_kills=0, page_spills=0;

///////////////////////////////////////////////////////////////////////
// cutil.h, (c) 1993-2006 NVIDIA Corporation.  All rights reserved.  //                                                           //
///////////////////////////////////////////////////////////////////////

#  define CMM_SAFE_CALL_NO_SYNC( call) do {                                 \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)

#  define CMM_SAFE_CALL( call) do {                                         \
    CMM_SAFE_CALL_NO_SYNC(call);                                            \
    cudaError err = cudaThreadSynchronize();                                 \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } } while (0)

#  define CMM_DEVICE_INIT() do {                                             \
    int deviceCount;                                                         \
    CMM_SAFE_CALL_NO_SYNC(cudaGetDeviceCount(&deviceCount));                \
    if (deviceCount == 0) {                                                  \
        fprintf(stderr, "There is no device.\n");                            \
        exit(EXIT_FAILURE);                                                  \
    }                                                                        \
    int dev;                                                                 \
    for (dev = 0; dev < deviceCount; ++dev) {                                \
        cudaDeviceProp deviceProp;                                           \
        CMM_SAFE_CALL_NO_SYNC(cudaGetDeviceProperties(&deviceProp, dev));   \
        if (deviceProp.major >= 1)                                           \
            break;                                                           \
    }                                                                        \
    if (dev == deviceCount) {                                                \
        fprintf(stderr, "There is no device supporting CUDA.\n");            \
        exit(EXIT_FAILURE);                                                  \
    }                                                                        \
    else                                                                     \
        CMM_SAFE_CALL(cudaSetDevice(dev));                                  \
} while (0)

///////////////////////////////////////////////////////////////////////
// cmmBlock                                                          //
///////////////////////////////////////////////////////////////////////
cmmBlock::cmmBlock(int xs, int xe, int ys, int ye,
                   int zs, int ze, int ws, int we)
{
  if (xs || xe)
  {
    m_start.push_back(xs);
    m_end.push_back(xe);
    m_size = xe-xs;
    if (ys || ye)
    {
      m_start.push_back(ys);
      m_end.push_back(ye);
      m_size *= ye-ys;
      if (zs || ze)
      {
        m_start.push_back(zs);
        m_end.push_back(ze);
        m_size *= ze-zs;
        if (ws || we)
        {
          m_start.push_back(ws);
          m_end.push_back(we);
          m_size *= we-ws;
        }
      }
    }
  }
  else
    m_size = 0;

  for (int ii=m_start.size(); ii < 4; ii++)
  {
    m_start.push_back(0);
    m_end.push_back(1);
  }
}

cmmBlock::cmmBlock(const cmmDims &dims)
{
  m_size = 1;
  for (int ii=0; ii < dims.Dims(); ii++)
  {
    int dim = dims.Dim(ii);
    m_start.push_back(0);
    m_end.push_back(dim);
    m_size *= dim;
  }
}

bool cmmBlock::operator==(const cmmBlock &rhs) const 
{
  return (m_start == rhs.m_start && m_end == rhs.m_end);
}

bool cmmBlock::operator!=(const cmmBlock &rhs) const 
{
  return (m_start != rhs.m_start || m_end != rhs.m_end);
}

int cmmBlock::Start(int d) const 
{
  return m_start[d];
}
int cmmBlock::End(int d) const 
{
  return m_end[d];
}
int cmmBlock::Size() const 
{
  return m_size;
}
int cmmBlock::Dims() const 
{
  return m_start.size();
}

///////////////////////////////////////////////////////////////////////
// cmmDims                                                           //
///////////////////////////////////////////////////////////////////////
cmmDims::cmmDims(int x, int y, int z, int w)
{
  if (x)
  {
    m_dims.push_back(x);
    m_size = x;
    if (y)
    {
      m_dims.push_back(y);
      m_size *= y;
      if (z)
      {
        m_dims.push_back(z);
        m_size *= z;
        if (w)
        {
          m_dims.push_back(w);
          m_size *= w;
        }
      }
    }
  }
  else
    m_size = 0;

  for (int ii=m_dims.size(); ii < 4; ii++)
    m_dims.push_back(1);
}

cmmDims::cmmDims(const cmmBlock &block)
{
  m_size = 1;
  for (int ii=0; ii < block.Dims(); ii++)
  {
    int dim = block.End(ii) - block.Start(ii);
    m_dims.push_back(dim);
    m_size *= dim;
  }
}

bool cmmDims::operator==(const cmmDims &rhs) const
{
  return m_dims == rhs.m_dims;
}

bool cmmDims::operator!=(const cmmDims &rhs) const
{
  return m_dims != rhs.m_dims;
}

int cmmDims::Dim(int d) const 
{
  return m_dims[d];
}

int cmmDims::Size() const 
{
  return m_size;
}

int cmmDims::Dims() const 
{
  return m_dims.size();
}

///////////////////////////////////////////////////////////////////////
// cmmState                                                          //
///////////////////////////////////////////////////////////////////////

cmmState::cmmState(cmmPageList *list)
{
  m_page = NULL;
  m_list = list;
}

cmmState::cmmState(cmmPage *page, cmmState &init)
{
  m_page = page;
  m_list = NULL;

  Change(init);
}

cmmState::~cmmState()
{
  if (m_page)
  {
    if (m_list)
      m_list->erase(m_position);
  }
  else
  {
    delete m_list;
  }
}

void cmmState::Change(cmmState &to)
{
  if (!m_page)
    throw runtime_error("Invalid object");

  if (m_list)
    m_list->erase(m_position);

  m_list = to.m_list;
  m_list->push_back(m_page);
  m_position = --m_list->end();
}

cmmStateStatistics cmmState::GetStatistics(void)
{
  cmmStateStatistics stats;

  for (PI pi=m_list->begin(); pi != m_list->end(); ++pi)
  {
    stats.pages++;
    if ((*pi)->m_devptr)
      stats.device_mem += (*pi)->m_block.Size()*sizeof(float);
    if ((*pi)->m_hostptr)
    {
      if ((*pi)->m_free)
        stats.host_mem_free += (*pi)->m_dims.Size()*sizeof(float);
      else
        stats.host_mem_nofree += (*pi)->m_dims.Size()*sizeof(float);
    }
  }

  return stats;
}

///////////////////////////////////////////////////////////////////////
// cmmPage                                                           //
///////////////////////////////////////////////////////////////////////

cmmPage::cmmPage() : m_state(cmmState(this, cmmStateNascent))
{
  if (!initialized)
  {
    CMM_DEVICE_INIT();
    initialized = true;
  }

  m_locks = 0;
  m_hostptr = m_devptr = NULL;
}

cmmPage::~cmmPage()
{
  if (m_state != cmmStateZombie)
    throw runtime_error("Invalid state");
}

void cmmPage::Associate(const cmmDims &dims, float *hostptr, bool free)
{
  if (m_state != cmmStateNascent)
    throw runtime_error("Invalid state");

  m_hostptr = hostptr;
  m_dims = dims;
  m_free = free;

  m_state.Change(cmmStateStored);
}

void cmmPage::Lock(const cmmBlock &block)
{
  bool allocated = false;

  page_locks++;

  if (m_state == cmmStateLocked && m_block == block)
  {
    m_locks++;
    return;
  }

  if (m_state == cmmStateCached)
  {
    if (m_block == block)
    {
      page_relocks++;
      m_locks++;
      m_state.Change(cmmStateLocked);
      return;
    }
    else
      Flush();
  }

  if (m_state != cmmStateStored)
    throw runtime_error("Invalid state");

  /* (1) Try to find a zombie page to take over */
  for (PI pi=cmmStateZombie.m_list->begin(); pi != cmmStateZombie.m_list->end(); ++pi)
  {
    if ((*pi)->m_block.Size() == block.Size())
    {
      page_hits++;
      m_devptr = (*pi)->m_devptr;
      delete (*pi);

      allocated = true;
      break;
    }
  }

  /* (2) Try to allocate */
  while (!allocated && cudaMalloc((void**)&m_devptr, block.Size()*sizeof(float)) != cudaSuccess)
  {
    /* (3) Free some zombie page */
    if (!cmmStateZombie.m_list->empty())
    {
      page_kills++;
      PI pi = cmmStateZombie.m_list->begin();
      CMM_SAFE_CALL(cudaFree((*pi)->m_devptr));
      (*pi)->m_devptr = NULL;
      delete (*pi);
    }
    /* (4) Flush some cached page */
    else if (!cmmStateCached.m_list->empty())
    {
      page_spills++;
      PI pi = cmmStateCached.m_list->begin();
      (*pi)->Flush();

      /* Check if the generated zombie was our size */
      pi = cmmStateZombie.m_list->begin();
      if ((*pi)->m_block.Size() == block.Size())
      {
        m_devptr = (*pi)->m_devptr;
        delete (*pi);

        allocated = true;
        break;
      }
    }
    else
      throw runtime_error("Out of memory");
  }

  // Clear out of memory error
  cudaGetLastError();

  if (m_hostptr)
  {
    if (block.Size() == m_dims.Size())
    {
      CMM_SAFE_CALL(cudaMemcpy(m_devptr, m_hostptr, m_dims.Size()*sizeof(float), cudaMemcpyHostToDevice));
    }
    else
    {
      // *** TODO: More than 2 dimensions
      CMM_SAFE_CALL(cudaMemcpy2D(m_devptr, 
                  (block.End(0)-block.Start(0))*sizeof(float),
                  m_hostptr+(block.Start(1)*m_dims.Dim(0)+block.Start(0)),
                  m_dims.Dim(0)*sizeof(float),
                  (block.End(0)-block.Start(0))*sizeof(float),
                  block.End(1)-block.Start(1),
                  cudaMemcpyHostToDevice));
    }
  }

  m_block = block;

  m_locks++;
  m_state.Change(cmmStateLocked);
}

void cmmPage::Release()
{
  if (m_state != cmmStateLocked)
    throw runtime_error("Invalid state");

  m_locks--;

  if (m_locks)
    return;

  m_state.Change(cmmStateCached);
}

void cmmPage::Flush()
{
  if (!m_hostptr)
    m_hostptr = (float*)malloc(m_dims.Size()*sizeof(float));

  if (m_state == cmmStateStored)
    return;

  if (m_state != cmmStateCached)
    throw runtime_error("Invalid state");

  if (m_block.Size() == m_dims.Size())
  {
    CMM_SAFE_CALL(cudaMemcpy(m_hostptr, m_devptr, m_dims.Size()*sizeof(float), cudaMemcpyDeviceToHost));
  }
  else
  {
    // *** TODO: More than 2 dimensions
    CMM_SAFE_CALL(cudaMemcpy2D(m_hostptr+(m_block.Start(1)*m_dims.Dim(0)+m_block.Start(0)),
                 m_dims.Dim(0)*sizeof(float),
                 m_devptr, 
                 (m_block.End(0)-m_block.Start(0))*sizeof(float),
                 (m_block.End(0)-m_block.Start(0))*sizeof(float),
                 m_block.End(1)-m_block.Start(1),
                 cudaMemcpyDeviceToHost));
  }

  // Put the memory in cold storage
  cmmPage *cache = new cmmPage();
  cache->m_block = m_block;
  cache->m_devptr = m_devptr;
  cache->m_state.Change(cmmStateZombie);

  m_devptr = NULL;
  m_block = cmmBlock();

  m_state.Change(cmmStateStored);
}

void cmmPage::Free()
{
  if (m_state != cmmStateStored && m_state != cmmStateCached)
    throw runtime_error("Invalid state");

  if (m_free && m_hostptr)
    free(m_hostptr);
  m_hostptr = NULL;

  m_state.Change(cmmStateZombie);
  if (!m_devptr)
    delete this;
}

///////////////////////////////////////////////////////////////////////
// cmmPagedMemory                                                    //
///////////////////////////////////////////////////////////////////////

cmmPagedMemory::cmmPagedMemory(const cmmDims &dims, float *hostptr, bool free)
{
  m_page = new cmmPage();
  m_page->Associate(dims, hostptr, free);
}

cmmPagedMemory::~cmmPagedMemory()
{
  m_page->Free();
}

float *cmmPagedMemory::Lock()
{
  return Lock(cmmBlock(GetDimensions()));
}

float *cmmPagedMemory::Lock(const cmmBlock &block)
{
  m_page->Lock(block);
  return m_page->m_devptr;
}

void cmmPagedMemory::Release()
{
  m_page->Release();
}

void cmmPagedMemory::Flush()
{
  m_page->Flush();
}

float *cmmPagedMemory::GetHostPtr(void)
{
  m_page->Flush();
  return m_page->m_hostptr;
}

cmmDims cmmPagedMemory::GetDimensions(void) const
{
  return m_page->m_dims;
}

float *cmmPagedMemory::GetDevicePtr(void) const
{
  return m_page->m_devptr;
}

cmmBlock cmmPagedMemory::GetBlock(void) const
{
  return m_page->m_block;
}

///////////////////////////////////////////////////////////////////////
// cmmStatistics                                                     //
///////////////////////////////////////////////////////////////////////

DLLIMPORTEXPORT cmmStatistics cmmGetStatistics(void)
{
  cmmStatistics stats;

  stats.stored = cmmStateStored.GetStatistics();
  stats.cached = cmmStateCached.GetStatistics();
  stats.locked = cmmStateLocked.GetStatistics();
  stats.zombie = cmmStateZombie.GetStatistics();

  stats.pages = stats.stored.pages+stats.cached.pages+stats.locked.pages+stats.zombie.pages;
  stats.device_mem = stats.stored.device_mem+stats.cached.device_mem+stats.locked.device_mem+stats.zombie.device_mem;
  stats.host_mem_free = stats.stored.host_mem_free+stats.cached.host_mem_free+stats.locked.host_mem_free+stats.zombie.host_mem_free;
  stats.host_mem_nofree = stats.stored.host_mem_nofree+stats.cached.host_mem_nofree+stats.locked.host_mem_nofree+stats.zombie.host_mem_nofree;

  stats.locks = page_locks;
  stats.relocks = page_relocks;
  stats.hits = page_hits;
  stats.kills = page_kills;
  stats.spills = page_spills;

  return stats;
}

char *cmmStatistics::Report(void)
{
  static char buf[2048];

  sprintf(buf, "Managing %d pages for %d MB device memory, %d MB managed host memory, and %d MB unmanaged host memory.\nDistribution:\n",
    pages, device_mem/1048576, host_mem_free/1048576, host_mem_nofree/1048576);
  sprintf(&buf[strlen(buf)], "  Stored: %d pages, %d MB, %d MB, %d MB\n",
    stored.pages, stored.device_mem/1048576, stored.host_mem_free/1048576, stored.host_mem_nofree/1048576);
  sprintf(&buf[strlen(buf)], "  Cached: %d pages, %d MB, %d MB, %d MB\n",
    cached.pages, cached.device_mem/1048576, cached.host_mem_free/1048576, cached.host_mem_nofree/1048576);
  sprintf(&buf[strlen(buf)], "  Locked: %d pages, %d MB, %d MB, %d MB\n",
    locked.pages, locked.device_mem/1048576, locked.host_mem_free/1048576, locked.host_mem_nofree/1048576);
  sprintf(&buf[strlen(buf)], "  Zombie: %d pages, %d MB, %d MB, %d MB\n",
    zombie.pages, zombie.device_mem/1048576, zombie.host_mem_free/1048576, zombie.host_mem_nofree/1048576);
  if (locks)
  {
    sprintf(&buf[strlen(buf)], "There were %d page locks, %d%% of which were relocks, and %d%% hit in the cache\n",
      locks, 100*relocks/locks, 100*hits/locks);
    sprintf(&buf[strlen(buf)], "%d zombie pages had to be put down, and %d pages spilled to free up memory\n",
      kills, spills);
  }

  unsigned int free, total;

  cuMemGetInfo(&free, &total);
  sprintf(&buf[strlen(buf)], "CUDA reports %d MB out of %d MB available\n", free/1048576, total/1048576);

  return buf;
}
